import xml.etree.ElementTree as et
from pathlib import Path

from django import forms
from django.core.exceptions import ValidationError
from django.core.validators import (
    FileExtensionValidator,
    get_available_image_extensions,
)
from django.db import models
from django.db.models.fields.files import FieldFile


def validate_svg(f):
    f.seek(0)
    tag = None
    try:
        for event, el in et.iterparse(f, ("start",)):
            tag = el.tag
            break
    except et.ParseError:
        pass

    if tag != "{http://www.w3.org/2000/svg}svg":
        raise ValidationError("Uploaded file is not an image or SVG file.")

    f.seek(0)

    return f


class SVGAndImageFieldFormField(forms.ImageField):
    default_validators = [
        FileExtensionValidator(
            allowed_extensions=get_available_image_extensions() + ["svg"]
        )
    ]

    def to_python(self, data):
        try:
            f = super().to_python(data)
        except ValidationError:
            return validate_svg(data)

        return f


class SVGAndImageField(models.ImageField):
    def formfield(self, **kwargs):
        defaults = {"form_class": SVGAndImageFieldFormField}
        defaults.update(kwargs)
        return super().formfield(**defaults)


class FieldVideo(FieldFile):
    @property
    def file_extension(self):
        return Path(self.name).suffix[1:]


class VideoField(models.FileField):
    attr_class = FieldVideo

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.validators.append(
            FileExtensionValidator(allowed_extensions=["mp4", "webm", "ogg"])
        )



import logging

from django.apps import AppConfig

logger = logging.getLogger(__name__)


class ContentBlocksConfig(AppConfig):
    name = "content_blocks"
    verbose_name = "Content blocks"

    def ready(self):
        from content_blocks.signals import (  # noqa
            cleanup_media_delete,
            cleanup_media_save,
        )



"""
Content blocks template tags.
"""
from django import template

from content_blocks.models import ContentBlockCollection
from content_blocks.services.content_block import RenderServices

register = template.Library()


@register.inclusion_tag(
    "content_blocks/content_block_collection.html", takes_context=True
)
def content_block_collection(context, content_block_collection_slug, **extra_context):
    """
    Render content blocks for a ContentBlockCollection given the collection's slug.
    """
    context.update({"slug": content_block_collection_slug})
    context.update(extra_context)

    try:
        collection = ContentBlockCollection.objects.get(
            slug=content_block_collection_slug
        )
        context.update({"content_block_collection": collection})
    except ContentBlockCollection.DoesNotExist:
        pass

    return context.flatten()


@register.simple_tag(takes_context=True)
def render_content_block(context, content_block, **extra_context):
    """
    Allows rendering of content blocks with request context.
    """
    context.update(extra_context)
    return RenderServices.render_content_block(content_block, context=context.flatten())


# todo render_content_blocks template tag to render a set of blocks given the parent

# todo render_content_block_previews template tag as above but renders previews



from django.core.management import BaseCommand

from content_blocks.models import ContentBlockTemplate, ContentBlockTemplateField
from content_blocks.services.content_block_template import ImportExportServices


class Command(BaseCommand):
    """
    Extension of loaddata command. Any ContentBlockTemplate specified in the import json is created or updated to
    match the import data.  For updated objects this includes deleting any related ContentBlockTemplateField which no
    longer exists and creating ContentBlockField objects for existing ContentBlock objects as required.
    """

    help = "Import content block templates from json file."

    # Track the ContentBlockTemplate and ContentBlockTemplateField objects that are imported.
    imported_pks = {
        ContentBlockTemplate: [],
        ContentBlockTemplateField: [],
    }

    infile = None

    def add_arguments(self, parser):
        """
        Add the infile argument, so we can pass file_like objects in call_command.
        """
        parser.add_argument(
            "args",
            metavar="filepath",
            nargs=1,
            help="Path to json file.",
        )

    def handle(self, filepath, **options):
        verbosity = int(options["verbosity"])

        ImportExportServices.import_content_block_templates_from_file(
            filepath, verbosity
        )

        if verbosity > 0:
            self.stdout.write("Content block templates imported!")



# Generated by Django 4.2.1 on 2023-07-12 02:23

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('content_blocks', '0003_alter_contentblocktemplatefield_unique_together_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='contentblocktemplate',
            name='no_cache',
        ),
    ]



# Generated by Django 4.2.6 on 2023-10-19 14:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('content_blocks', '0005_alter_contentblockcollection_options'),
    ]

    operations = [
        migrations.CreateModel(
            name='IframeField',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('content_blocks.contentblockfield',),
        ),
        migrations.AddField(
            model_name='contentblockfield',
            name='iframe',
            field=models.CharField(blank=True, max_length=256),
        ),
    ]



# Generated by Django 4.2.6 on 2023-10-19 17:29

from django.db import migrations


def populate_content_block_names(apps, schema_editor):
    ContentBlock = apps.get_model("content_blocks", "ContentBlock")

    for content_block in ContentBlock.objects.all():
        content_block.name = (
            f"{content_block.content_block_template.name} #{content_block.id}"
        )
        content_block.save()


class Migration(migrations.Migration):

    dependencies = [
        ("content_blocks", "0007_contentblock_name_and_more"),
    ]

    operations = [
        migrations.RunPython(populate_content_block_names, migrations.RunPython.noop)
    ]



"""
Content blocks app signals.py
"""
from django.db.models.signals import pre_delete, pre_save
from django.dispatch import Signal, receiver

from content_blocks.models import (
    ContentBlockField,
    ContentBlockFields,
    FileField,
    ImageField,
    VideoField,
)

# A signal we can send after an import finishes.
post_import = Signal()


def cleanup_media(sender, instance, delete=False, **kwargs):
    """
    Delete old media files when no longer needed.
    """
    if kwargs.get("raw", False):
        # Prevent this signal from running during loaddata.
        return

    if instance.id is None:
        return  # New object being created

    object_types = {
        ContentBlockFields.IMAGE_FIELD: "image",
        ContentBlockFields.FILE_FIELD: "file",
        ContentBlockFields.VIDEO_FIELD: "video",
    }

    object_type = object_types.get(instance.field_type)
    if object_type is None:
        return  # pragma: no cover

    if delete:
        old_file = getattr(instance, object_type)
        new_file = None
    else:
        try:
            old_file = getattr(
                ContentBlockField.objects.get(id=instance.id), object_type
            )
            new_file = getattr(instance, object_type)
        except ContentBlockField.DoesNotExist:  # pragma: no cover
            return

    if (
        old_file != new_file
        and not ContentBlockField.objects.filter(**{object_type: old_file})
        .exclude(id=instance.id)
        .exists()
    ):
        old_file.delete(save=False)


@receiver(pre_save, sender=ImageField, dispatch_uid="cleanup_image_media_save")
@receiver(pre_save, sender=FileField, dispatch_uid="cleanup_file_media_save")
@receiver(pre_save, sender=VideoField, dispatch_uid="cleanup_video_media_save")
@receiver(pre_save, sender=ContentBlockField, dispatch_uid="cleanup_media_save")
def cleanup_media_save(sender, instance, **kwargs):
    return cleanup_media(sender, instance, delete=False, **kwargs)


@receiver(pre_delete, sender=ImageField, dispatch_uid="cleanup_image_media_delete")
@receiver(pre_delete, sender=FileField, dispatch_uid="cleanup_file_media_delete")
@receiver(pre_delete, sender=VideoField, dispatch_uid="cleanup_video_media_delete")
@receiver(pre_delete, sender=ContentBlockField, dispatch_uid="cleanup_media_delete")
def cleanup_media_delete(sender, instance, **kwargs):
    return cleanup_media(sender, instance, delete=True, **kwargs)



# Generated by Django 4.2.1 on 2023-09-04 22:06

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('content_blocks', '0004_remove_contentblocktemplate_no_cache'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='contentblockcollection',
            options={'ordering': ['name']},
        ),
    ]



from django.template.loader import render_to_string

from content_blocks.models import ContentBlockFields


class ContentBlockFilters:
    """
    ContentBlock queryset filters.
    """


class RenderServices:
    """
    Services for rendering ContentBlock to html.
    """

    @staticmethod
    def render_content_block(content_block, context=None):
        """
        Render the html for the given ContentBlock.
        :context: Dictionary of context to render the template with.
        :return: Rendered html for the content block.
        """
        if not content_block.can_render:
            return ""

        context = RenderServices.context(content_block, context=context)
        html = render_to_string(content_block.template, context)
        return html

    @staticmethod
    def context(content_block, context=None):
        """
        Adds the ``ContentBlock.context`` and ``ContentBlock`` to supplied context or creates new context.
        :return: Context dictionary to be used when rendering html.
        """
        context = context or {}
        context[content_block.context_name] = content_block.context
        context[f"{content_block.context_name}_object"] = content_block
        return context


class CloneServices:
    """
    Services for cloning ContentBlock.
    """

    @staticmethod
    def clone_content_block(content_block, attrs=None):
        """
        Clones the given content block and all content block fields.
        """
        new_content_block = content_block.make_clone(attrs=attrs)

        for field in content_block.content_block_fields.all():
            new_field = field.make_clone(attrs={"content_block": new_content_block})

            if field.template_field.field_type == ContentBlockFields.NESTED_FIELD:
                for nested_block in field.content_blocks.all():
                    CloneServices.clone_content_block(
                        nested_block, attrs={"parent": new_field}
                    )

        return new_content_block



# Generated by Django 4.2.6 on 2023-10-19 19:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('content_blocks', '0008_auto_20231019_1729'),
    ]

    operations = [
        migrations.AlterField(
            model_name='contentblock',
            name='name',
            field=models.CharField(blank=True, help_text='Used to identify content blocks in the content block editor. Defaults to <content_block_template_name> #<content_block_id>', max_length=320),
        ),
    ]



from django import template

from content_blocks.forms import ContentBlockForm, NewNestedBlockForm

register = template.Library()


@register.simple_tag
def content_block_form(content_block):
    return ContentBlockForm(content_block=content_block)


@register.simple_tag
def new_nested_block_form(parent):
    return NewNestedBlockForm(
        initial={
            "parent": parent,
        }
    )


@register.simple_tag
def app_model_label(obj):
    return f"{obj._meta.app_label}.{obj._meta.model.__name__}"



from django import forms
from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.db import transaction

from content_blocks.conf import settings
from content_blocks.models import (
    ContentBlock,
    ContentBlockAvailability,
    ContentBlockField,
    ContentBlockFields,
    ContentBlockTemplate,
)
from content_blocks.services.content_block import CloneServices, RenderServices

if apps.is_installed("django.contrib.sites"):
    from django.contrib.sites.models import Site


class ParentModelForm(forms.Form):
    """
    Adds parent object to form from kwargs.
    """

    parent = None

    def __init__(self, *args, **kwargs):
        self.parent = kwargs.pop("parent")
        super().__init__(*args, **kwargs)


class NewContentBlockFormBase(forms.Form):
    """
    Shared functionality between NewContentBlockForm and NewNestedBlockForm.
    Has a select for content block template.
    """

    content_block_template = forms.ModelChoiceField(
        ContentBlockTemplate.objects.visible(), empty_label=None
    )

    position = forms.CharField(initial=0, widget=forms.HiddenInput())

    auto_id = True

    def __init__(self, *args, **kwargs):
        kwargs["auto_id"] = self.auto_id
        super().__init__(*args, **kwargs)
        self.fields["content_block_template"].queryset = self.get_available_templates()

    def get_available_templates(self):
        """
        Must be provided by subclasses.
        """
        raise NotImplementedError  # pragma: no cover

    def create_content_block(self, content_block_template, draft=False, **kwargs):
        kwargs["position"] = kwargs.get(
            "position", self.cleaned_data.get("position", 0)
        )
        # todo refactor to service class
        with transaction.atomic():
            content_block = ContentBlock.objects.create(
                content_block_template=content_block_template,
                draft=draft,
                **kwargs,
            )
            content_block.name = f"{content_block_template.name} #{content_block.id}"
            content_block.save()

            for (
                template_field
            ) in content_block_template.content_block_template_fields.all():
                content_block_field = ContentBlockField.objects.create(
                    content_block=content_block,
                    template_field=template_field,
                    field_type=template_field.field_type,
                    model_choice_content_type=template_field.model_choice_content_type,
                )

                # create min_num nested blocks
                if template_field.field_type == ContentBlockFields.NESTED_FIELD:
                    for j in range(template_field.min_num):
                        self.create_content_block(
                            template_field.nested_templates.first(),
                            draft=False,
                            parent=content_block_field,
                            position=j,
                        )

        return content_block


class NewContentBlockForm(ParentModelForm, NewContentBlockFormBase):
    """
    Form for creating new top level content blocks.
    Creates new content block object with appropriate fields on save.
    Updates the parent which the content block belongs to m2m on save.
    """

    auto_id = "new_cb_%s"

    def get_available_templates(self):
        # todo refactor to service class
        try:
            content_type = ContentType.objects.get_for_model(self.parent)
            content_block_availability = ContentBlockAvailability.objects.get(
                content_type=content_type
            )
            return content_block_availability.content_block_templates.visible()
        except ContentBlockAvailability.DoesNotExist:
            return ContentBlockTemplate.objects.visible()

    def update_parent_m2m(self, content_block):
        self.parent.content_blocks.add(content_block)

    def save(self):
        # todo call service class
        content_block = self.create_content_block(
            self.cleaned_data["content_block_template"], draft=True
        )
        self.update_parent_m2m(content_block)
        return content_block


class NewNestedBlockForm(NewContentBlockFormBase):
    """
    As per NewContentBlockForm but for adding nested blocks.
    """

    parent = forms.ModelChoiceField(
        widget=forms.HiddenInput(), queryset=ContentBlockField.objects.all()
    )

    auto_id = False

    def get_available_templates(self):
        # todo refactor to service class
        try:
            parent = self.initial["parent"]
            return parent.template_field.nested_templates.visible()
        except KeyError:
            return ContentBlockTemplate.objects.all()

    def save(self):
        # todo call service class
        return self.create_content_block(
            self.cleaned_data["content_block_template"],
            draft=False,
            parent=self.cleaned_data["parent"],
        )


class ContentBlockForm(forms.Form):
    """
    Form for creating and changing content blocks.
    Form fields are generated from the content block fields.
    On save each content block field is updated and saved.
    """

    name = forms.CharField(
        max_length=320,
    )

    css_class = forms.CharField(
        max_length=64,
        required=False,
        label="CSS Class",
        help_text="For advanced styling.",
    )

    def __init__(self, *args, **kwargs):
        self.content_block = kwargs.pop("content_block")
        kwargs["auto_id"] = f"id_%s_{self.content_block.id}"
        super().__init__(*args, **kwargs)
        self.set_fields()
        self.fields["css_class"].initial = self.content_block.css_class
        self.fields["name"].initial = self.content_block.name

    def set_fields(self):
        for key, field in self.content_block.fields.items():
            form_field = field.form_field
            if form_field:
                self.fields[key] = form_field

    def save(self):
        # todo refactor to service class?
        with transaction.atomic():
            for key, field in self.content_block.fields.items():
                if key in self.cleaned_data.keys():
                    field.save_value(self.cleaned_data.get(key))

            self.content_block.css_class = self.cleaned_data.get("css_class")
            self.content_block.name = self.cleaned_data.get("name")
            self.content_block.saved = True
            self.content_block.save()

            # We no longer manage the cache here as drafts and nested blocks are not cached.

        return self.content_block


class PublishContentBlocksForm(ParentModelForm):
    """
    Form for publishing content blocks.
    Duplicates all content blocks and fields and nested blocks and sets draft=False
    """

    def save(self):
        # todo refactor to service class
        with transaction.atomic():
            sites = (
                list(Site.objects.all())
                if apps.is_installed("django.contrib.sites")
                else [None]
            )
            self.parent.content_blocks.published().delete()

            for content_block in self.parent.content_blocks.drafts():
                new_content_block = CloneServices.clone_content_block(
                    content_block, attrs={"draft": False}
                )
                self.parent.content_blocks.add(new_content_block)

                if settings.CONTENT_BLOCKS_PRE_RENDER:
                    for site in sites:
                        RenderServices.render_content_block(
                            new_content_block,
                            context={
                                "cache_timeout": settings.CONTENT_BLOCKS_PRE_RENDER_CACHE_TIMEOUT,
                                "site": site,
                            },
                        )


class ResetContentBlocksForm(ParentModelForm):
    """
    Deletes existing draft content blocks and creates new from published content blocks.
    """

    def save(self):
        # todo refactor to service class
        with transaction.atomic():
            self.parent.content_blocks.drafts().delete()

            for content_block in self.parent.content_blocks.published():
                new_content_block = CloneServices.clone_content_block(
                    content_block, attrs={"draft": True}
                )
                self.parent.content_blocks.add(new_content_block)


class ImportContentBlocksForm(ParentModelForm):
    """
    Import content blocks from another master object of the same type as parent.
    """

    master = forms.ModelChoiceField(None, empty_label=None)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["master"].queryset = self.get_master_queryset()

    def get_master_queryset(self):
        return self.parent._meta.model.objects.exclude(id=self.parent.id)

    def save(self):
        # todo refactor to service class
        with transaction.atomic():
            self.parent.content_blocks.drafts().delete()

            for content_block in self.cleaned_data["master"].content_blocks.drafts():
                new_content_block = CloneServices.clone_content_block(content_block)
                self.parent.content_blocks.add(new_content_block)



"""
Content blocks admin.py
"""
from io import StringIO

from adminsortable2.admin import SortableAdminMixin, SortableInlineAdminMixin
from django.apps import apps
from django.contrib import admin, messages
from django.http import HttpResponseRedirect, StreamingHttpResponse
from django.template import TemplateDoesNotExist
from django.template.loader import get_template
from django.urls import path, reverse
from django.utils.safestring import mark_safe

from content_blocks.admin_forms import (
    ContentBlockTemplateAdminForm,
    ContentBlockTemplateFieldAdminForm,
    ContentBlockTemplateFieldInlineFormSet,
)
from content_blocks.models import (
    ContentBlockAvailability,
    ContentBlockCollection,
    ContentBlockTemplate,
    ContentBlockTemplateField,
)
from content_blocks.services.content_block_template import ImportExportServices
from content_blocks.views import (
    content_block_create,
    content_block_editor,
    content_block_preview,
    content_block_template_export,
    content_block_template_import,
    discard_changes,
    import_content_blocks,
    nested_block_create,
    publish_content_blocks,
)

#  Admin constants and base classes

# Auto dates
AUTO_DATE_FIELDS = ["create_date", "mod_date"]
AUTO_DATE_FIELDSET = (
    "Dates",
    {"fields": (("create_date", "mod_date"),)},
)

# Visible
VISIBLE_FIELDS = ["visible"]
VISIBLE_FIELDSET = ("Visibility", {"fields": VISIBLE_FIELDS})


class ContentBlockTemplateFieldInline(SortableInlineAdminMixin, admin.StackedInline):
    model = ContentBlockTemplateField
    fk_name = "content_block_template"
    formset = ContentBlockTemplateFieldInlineFormSet
    form = ContentBlockTemplateFieldAdminForm
    min_num = 0
    extra = 0
    autocomplete_fields = ["nested_templates"]


@admin.register(ContentBlockTemplate)
class ContentBlockTemplateAdmin(SortableAdminMixin, admin.ModelAdmin):
    form = ContentBlockTemplateAdminForm

    change_list_template = (
        "content_blocks/admin/content_block_template_change_list.html"
    )
    list_display = [
        "name",
        "template_filename",
        "visible",
    ] + AUTO_DATE_FIELDS
    list_editable = ("visible",)
    list_filter = AUTO_DATE_FIELDS + ["visible"]
    search_fields = ("name", "template_filename")

    readonly_fields = AUTO_DATE_FIELDS
    inlines = [ContentBlockTemplateFieldInline]
    actions = ["export_content_block_templates"]
    save_as = True

    class Media:
        css = {"all": ["content_blocks/css/content_block_template_admin.css"]}
        js = (
            "content_blocks/js/content_block_template_admin.js",
            "content_blocks/js/admin_choices_widget.js",
        )

    @property
    def content_block_template_fields(self):
        return ["name", "template_filename"]

    def get_fieldsets(self, request, obj=None):
        return (
            AUTO_DATE_FIELDSET,
            VISIBLE_FIELDSET,
            (
                "Content Block Template",
                {
                    "fields": self.content_block_template_fields,
                },
            ),
        )

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)

        template_filename = form.cleaned_data.get("template_filename")

        if not template_filename:
            return

        template_filename = f"content_blocks/content_blocks/{template_filename}"

        try:
            get_template(template_filename)
        except TemplateDoesNotExist:
            url = reverse(
                "admin:content_blocks_contentblocktemplate_change", args=[obj.id]
            )
            msg = (
                f"Couldn't load a template for the content block template "
                f'"<a href="{url}">{obj}</a>". '
                f'No template found at "{template_filename}".'
            )
            messages.warning(request, mark_safe(msg))

    def get_urls(self):
        urls = [
            path(
                "export/",
                content_block_template_export,
                name="content_blocks_contentblocktemplate_export",
            ),
            path(
                "import/",
                content_block_template_import,
                {"model_admin": self},
                name="content_blocks_contentblocktemplate_import",
            ),
        ]
        return urls + super().get_urls()

    @admin.action(description="Export selected content block templates")
    def export_content_block_templates(self, request, queryset):
        """
        Export the selected ContentBlockTemplate objects as JSON suitable for import.
        """
        buffer = StringIO()
        ImportExportServices.export_content_block_templates(queryset, buffer)
        buffer.seek(0)

        return StreamingHttpResponse(
            buffer,
            content_type="application/json",
            headers={
                "Content-Disposition": 'attachment; filename="content_block_templates.json"'
            },
        )


if apps.is_installed("dbtemplates"):
    from dbtemplates.models import Template

    admin.site.unregister(ContentBlockTemplate)

    @admin.register(ContentBlockTemplate)
    class ContentBlockTemplateDBTemplatesAdmin(ContentBlockTemplateAdmin):
        @property
        def content_block_template_fields(self):
            return super().content_block_template_fields + ["db_template_button"]

        @admin.display(description="HTML Template")
        def db_template_button(self, obj):
            if not obj.template_filename:
                return "-"
            button = '<input type="submit" value="Save and edit HTML template" name="_dbtemplate">'
            return mark_safe(button)

        def get_readonly_fields(self, request, obj=None):
            return [
                *super().get_readonly_fields(request, obj=obj),
                "db_template_button",
            ]

        def _response(self, response, request, obj):
            if "_dbtemplate" in request.POST and obj.template:
                db_template, _ = Template.objects.get_or_create(name=obj.template)
                url = reverse(
                    "admin:dbtemplates_template_change", args=[db_template.id]
                )
                return HttpResponseRedirect(url)

            return response

        def response_change(self, request, obj):
            response = super().response_change(request, obj)
            return self._response(response, request, obj)

        def response_add(self, request, obj, post_url_continue=None):
            response = super().response_add(
                request, obj, post_url_continue=post_url_continue
            )
            return self._response(response, request, obj)


@admin.register(ContentBlockAvailability)
class ContentBlockAvailabilityAdmin(admin.ModelAdmin):
    list_display = ["content_type"] + AUTO_DATE_FIELDS
    list_filter = AUTO_DATE_FIELDS

    readonly_fields = AUTO_DATE_FIELDS
    filter_horizontal = ["content_block_templates"]

    fieldsets = (
        AUTO_DATE_FIELDSET,
        (
            "Content Block Availability",
            {"fields": ("content_type", "content_block_templates")},
        ),
    )


class ContentBlockModelAdmin(admin.ModelAdmin):
    """
    Base class to be added to the admin of any model which has a content_blocks m2m.  This will then add
    content_blocks_button which can then be added to the formset.
    """

    @admin.display(description="Content blocks")
    def content_blocks_link(self, obj):
        """
        A link to the content block editor for this object.
        Intended to be used on admin changelist.
        :param obj: subclass of :py:class:`ContentBlockParentModel`.
        """
        app_label, model_name = self.model._meta.app_label, self.model._meta.model_name
        url = reverse(
            f"admin:{app_label}_{model_name}_content_block_editor", args=[obj.id]
        )
        link = f'<a href="{url}">Edit content blocks</a>'
        return mark_safe(link)

    def content_blocks_button(self, *args):
        """
        A button which will save the object then, if successful, will redirect to the content block editor for this
        object.
        Intended to be used on the admin changepage.
        """
        button = '<input type="submit" value="Save and edit content blocks" name="_contentblocks">'
        return mark_safe(button)

    content_blocks_button.short_description = "Content blocks"

    def get_readonly_fields(self, *args, **kwargs):
        return (
            *super().get_readonly_fields(*args, **kwargs),
            "content_blocks_link",
            "content_blocks_button",
        )

    def _response(self, response, request, obj):
        if "_contentblocks" in request.POST:
            model_name = self.model._meta.model_name
            app_label = self.model._meta.app_label
            url = reverse(
                f"admin:{app_label}_{model_name}_content_block_editor", args=[obj.id]
            )
            return HttpResponseRedirect(url)

        return response

    def response_change(self, request, obj):
        response = super().response_change(request, obj)
        return self._response(response, request, obj)

    def response_add(self, request, obj, post_url_continue=None):
        response = super().response_add(
            request, obj, post_url_continue=post_url_continue
        )
        return self._response(response, request, obj)

    def get_urls(self):
        app_label, model_name = self.model._meta.app_label, self.model._meta.model_name
        urls = [
            path(
                "<path:object_id>/content-blocks/",
                content_block_editor,
                {"model_admin": self},
                name=f"{app_label}_{model_name}_content_block_editor",
            ),
            path(
                "<path:object_id>/content-blocks/_ajax/content-block-create/",
                content_block_create,
                {"model_admin": self},
                name=f"{app_label}_{model_name}_content_block_create",
            ),
            path(
                "<path:object_id>/content-blocks/_ajax/nested-block-create/",
                nested_block_create,
                {"model_admin": self},
                name=f"{app_label}_{model_name}_nested_block_create",
            ),
            path(
                "<path:object_id>/content-blocks/_ajax/publish/",
                publish_content_blocks,
                {"model_admin": self},
                name=f"{app_label}_{model_name}_publish_content_blocks",
            ),
            path(
                "<path:object_id>/content-blocks/_ajax/discard-changes/",
                discard_changes,
                {"model_admin": self},
                name=f"{app_label}_{model_name}_discard_changes",
            ),
            path(
                "<path:object_id>/content-blocks/_ajax/import-content-blocks/",
                import_content_blocks,
                {"model_admin": self},
                name=f"{app_label}_{model_name}_import_content_blocks",
            ),
            path(
                "<path:object_id>/content-blocks/<int:content_block_id>/preview/",
                content_block_preview,
                {"model_admin": self},
                name=f"{app_label}_{model_name}_content_block_preview",
            ),
        ]
        return urls + super().get_urls()


@admin.register(ContentBlockCollection)
class ContentBlockCollectionAdmin(ContentBlockModelAdmin):
    list_display = ["name", "slug", "content_blocks_link"] + AUTO_DATE_FIELDS
    list_display_links = ["name", "slug"]
    list_filter = AUTO_DATE_FIELDS
    search_fields = ["name", "slug"]

    readonly_fields = AUTO_DATE_FIELDS
    prepopulated_fields = {"slug": ("name",)}

    fieldsets = (
        AUTO_DATE_FIELDSET,
        (
            "Content Block Collection",
            {"fields": ("name", "slug", "content_blocks_button")},
        ),
    )



import json

from adminsortable2.admin import CustomInlineFormSet
from django import forms
from django.core.exceptions import ValidationError
from django.core.validators import FileExtensionValidator

from content_blocks.models import (
    ContentBlock,
    ContentBlockField,
    ContentBlockFields,
    ContentBlockTemplate,
    ContentBlockTemplateField,
)
from content_blocks.services.content_block_template import ImportExportServices
from content_blocks.widgets import (
    ChoicesWidget,
    ContentBlockTemplateFieldDeleteWidget,
    TemplateFilenameAutocompleteWidget,
)

REQUIRED_ERROR_MSG = "This field is required."


class ContentBlockTemplateAdminForm(forms.ModelForm):
    class Meta:
        model = ContentBlockTemplate
        exclude = []
        widgets = {
            "template_filename": TemplateFilenameAutocompleteWidget(
                "/content_blocks/content_blocks"
            )
        }


def validate_choices(choices):
    try:
        choices = json.loads(choices)
    except json.JSONDecodeError:
        return False

    if not isinstance(choices, list):
        return False

    for choice in choices:
        if not isinstance(choice, list):
            return False

        if not len(choice) == 2:
            return False

    return True


class ContentBlockTemplateFieldInlineFormSet(CustomInlineFormSet):
    deletion_widget = ContentBlockTemplateFieldDeleteWidget()


class ContentBlockTemplateFieldAdminForm(forms.ModelForm):
    class Meta:
        model = ContentBlockTemplateField
        exclude = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["choices"] = forms.CharField(
            widget=ChoicesWidget(),
            help_text="You must provide a value and label for each choice or it will be ignored.",
            required=False,
        )

    def clean(self):
        cleaned_data = super().clean()

        field_type = cleaned_data.get("field_type")

        if field_type == ContentBlockFields.NESTED_FIELD:
            nested_templates = cleaned_data.get("nested_templates")
            if not nested_templates:
                self.add_error("nested_templates", REQUIRED_ERROR_MSG)

            min_num = cleaned_data.get("min_num")
            max_num = cleaned_data.get("max_num")

            if min_num > max_num:
                self.add_error(
                    "min_num", "Min num must be less than or equal to max num."
                )
                self.add_error(
                    "max_num", "Max num must be greater than or equal to min num."
                )

        elif field_type == ContentBlockFields.CHOICE_FIELD:
            choices = cleaned_data.get("choices")
            if not choices:
                self.add_error("choices", REQUIRED_ERROR_MSG)
            elif not validate_choices(choices):
                # Validate choices
                self.add_error("choices", "Invalid choices")

        elif field_type == ContentBlockFields.MODEL_CHOICE_FIELD:
            model_choice_content_type = cleaned_data.get("model_choice_content_type")
            if not model_choice_content_type:
                self.add_error("model_choice_content_type", REQUIRED_ERROR_MSG)

        if self.instance.pk and "field_type" in self.changed_data:
            self.add_error("field_type", "Field type cannot be changed.")

        return cleaned_data

    def save(self, commit=True):
        """
        Update content block fields based on this template field.
        """
        template_field = super().save(commit=False)

        created = False if template_field.pk else True

        # Find existing content blocks using this field
        content_blocks = ContentBlock.objects.filter(
            content_block_template=template_field.content_block_template
        )
        create_content_block_fields = created and content_blocks.exists()

        if commit or create_content_block_fields:
            template_field.save()
            self.save_m2m()

        if create_content_block_fields:
            # Create a new field for existing content blocks
            for content_block in content_blocks:
                ContentBlockField.objects.create(
                    template_field=template_field,
                    content_block=content_block,
                    field_type=template_field.field_type,
                )

        return template_field


class ContentBlockTemplateImportForm(forms.Form):
    fixture_file = forms.FileField(
        validators=[FileExtensionValidator(allowed_extensions=["json"])]
    )

    def clean_fixture_file(self):
        """
        Validate the uploaded file is JSON.
        """
        fixture_file = self.cleaned_data["fixture_file"]
        value = fixture_file.read()

        try:
            json.loads(value)
        except json.JSONDecodeError:
            fixture_file.close()
            raise ValidationError("The file is not valid JSON.")

        return fixture_file

    def import_content_block_templates(self, fixture_file):
        """
        Only call this after testing is_valid()
        :param fixture_file: The file from request.FILES
        """
        with fixture_file.open() as file:
            ImportExportServices.import_content_block_templates(file)



# Generated by Django 4.2.6 on 2023-10-19 17:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('content_blocks', '0006_iframefield_contentblockfield_iframe'),
    ]

    operations = [
        migrations.AddField(
            model_name='contentblock',
            name='name',
            field=models.CharField(blank=True, help_text='Used to identify content blocks in the content block editor. Defaults to <content_block_template_name> #<content_block_id>', max_length=64),
        ),
        migrations.AlterField(
            model_name='contentblocktemplatefield',
            name='field_type',
            field=models.CharField(choices=[('TextField', 'Text Field'), ('ContentField', 'Content Field'), ('ImageField', 'Image Field'), ('NestedField', 'Nested Field'), ('CheckboxField', 'Checkbox Field'), ('ChoiceField', 'Choice Field'), ('ModelChoiceField', 'Model Choice Field'), ('FileField', 'File Field'), ('VideoField', 'Video Field'), ('EmbeddedVideoField', 'Embedded Video Field'), ('IframeField', 'Iframe Field')], max_length=32),
        ),
    ]



"""
Content Blocks models.py
"""
import functools
import json
import logging

from django import forms
from django.apps import apps
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.files.storage import get_storage_class
from django.core.validators import RegexValidator
from django.db import models
from django.forms.utils import pretty_name
from django.template.exceptions import TemplateDoesNotExist
from django.template.loader import get_template
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from model_clone import CloneMixin

from content_blocks.abstract_models import (
    AutoDateModel,
    PositionModel,
    VisibleManager,
    VisibleModel,
)
from content_blocks.conf import settings
from content_blocks.fields import (
    SVGAndImageField,
    SVGAndImageFieldFormField,
    VideoField,
)
from content_blocks.widgets import FileWidget

logger = logging.getLogger(__name__)

if apps.is_installed("django_cleanup") or apps.is_installed(
    "django_cleanup.apps.CleanupConfig"
):
    from django_cleanup.cleanup import cleanup_ignore
else:
    # noop shim for when django_cleanup isn't installed
    def cleanup_ignore(cls):  # pragma: no cover (covered in settings specific tests)
        return cls


class PolymorphError(Exception):
    """
    Raised should polymorph fail.
    """


class ContentBlockFields(models.TextChoices):
    TEXT_FIELD = "TextField"
    CONTENT_FIELD = "ContentField"
    IMAGE_FIELD = "ImageField"
    NESTED_FIELD = "NestedField"
    CHECKBOX_FIELD = "CheckboxField"
    CHOICE_FIELD = "ChoiceField"
    MODEL_CHOICE_FIELD = "ModelChoiceField"
    FILE_FIELD = "FileField"
    VIDEO_FIELD = "VideoField"
    EMBEDDED_VIDEO_FIELD = "EmbeddedVideoField"
    IFRAME_FIELD = "IframeField"


class ContentBlockFieldManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().select_related("template_field")


def get_storage(field_type):
    """
    Get the storage class from dotted strings in settings.
    If the field storage class is in settings use that otherwise use the general storage class setting for content
    blocks, which will default to STORAGES["default"].
    """
    storage_setting = getattr(settings, field_type, None)
    if storage_setting is not None:
        return get_storage_class(
            storage_setting
        )()  # pragma: no cover (covered by settings specific tests)
    return get_storage_class(settings.CONTENT_BLOCKS_STORAGE)()


def image_storage():
    return get_storage("CONTENT_BLOCKS_IMAGE_STORAGE")


def file_storage():
    return get_storage("CONTENT_BLOCKS_FILE_STORAGE")


def video_storage():
    return get_storage("CONTENT_BLOCKS_VIDEO_STORAGE")


@cleanup_ignore
class ContentBlockField(models.Model, CloneMixin):
    """
    Base model for all content block field models.
    """

    objects = ContentBlockFieldManager()
    template_field = models.ForeignKey(
        "content_blocks.ContentBlockTemplateField",
        on_delete=models.CASCADE,
        related_name="fields",
    )

    content_block = models.ForeignKey(
        "content_blocks.ContentBlock",
        on_delete=models.CASCADE,
        related_name="content_block_fields",
    )

    # duplicate from template_field here because it can't be changed on template field and seems impossible to have
    # select_related working in the __init__
    field_type = models.CharField(max_length=32)

    text = models.CharField(max_length=256, blank=True)
    content = models.TextField(blank=True)
    checkbox = models.BooleanField(blank=True, default=False)
    image = SVGAndImageField(
        upload_to="content-blocks/images", blank=True, storage=image_storage
    )
    file = models.FileField(
        upload_to="content-blocks/files", blank=True, storage=file_storage
    )
    choice = models.CharField(max_length=256, blank=True)
    video = VideoField(
        upload_to="content-blocks/videos", blank=True, storage=video_storage
    )
    embedded_video = models.CharField(max_length=256, blank=True)
    iframe = models.CharField(max_length=256, blank=True)

    model_choice_content_type = models.ForeignKey(
        ContentType, on_delete=models.SET_NULL, blank=True, null=True
    )
    model_choice_object_id = models.PositiveIntegerField(blank=True, null=True)
    model_choice = GenericForeignKey(
        "model_choice_content_type", "model_choice_object_id"
    )

    template_name = "content_blocks/partials/fields/default.html"
    preview_template_name = None

    class Meta:
        ordering = ["template_field__position"]

    def __init_subclass__(subcls):
        """
        Wrap all subclass form_field.__get__ methods with form_field_wrapper
        """
        super.__init_subclass__()

        new_property = property(
            ContentBlockField.form_field_wrapper(subcls.form_field.__get__),
            subcls.form_field.__set__,
        )
        setattr(
            subcls,
            "form_field",
            new_property,
        )

    def __getattribute__(self, item):
        """
        Lazy polymorphism.  Only polymorph when accessing certain attrs.
        """
        polymorph_attrs = [
            "form_field",
            "context_value",
            "save_value",
            "template_name",
            "preview_template_name",
            "label",
        ]
        if item in polymorph_attrs:
            self.polymorph()

        return super().__getattribute__(item)

    def polymorph(self):
        """
        Polymorph this model into a proxy model subclass.
        """
        if self.__class__ != ContentBlockField:
            # Don't polymorph more than once.
            return

        for subclass in ContentBlockField.__subclasses__():
            if self.field_type == subclass.__name__:
                self.__class__ = subclass
                return

        raise PolymorphError(f"{self} has no subclass {self.field_type}.")

    @staticmethod
    def form_field_wrapper(form_field):
        """
        Wraps the form_field.__get__ method and adds common attributes to all fields.
        :param form_field: the form_field.__get__ method we are wrapping
        """

        @functools.wraps(form_field)
        def wrapper(self):
            field = form_field(self)
            if field is not None:
                field.cb_field = self

            return field

        return wrapper

    @property
    def form_field(self):
        """
        To be overridden. Will return a form.Field class if the field is to be shown on the form.
        This method should return _form_field() with the desired forms.Field.
        """
        raise NotImplementedError  # pragma: no cover

    @property
    def context_value(self):
        raise NotImplementedError  # pragma: no cover

    def save_value(self, value):
        raise NotImplementedError  # pragma: no cover

    @property
    def key(self):
        return self.template_field.key


class TextField(ContentBlockField):
    class Meta:
        proxy = True

    @property
    def context_value(self):
        if settings.CONTENT_BLOCKS_MARK_SAFE:
            return mark_safe(self.text)
        return self.text

    def save_value(self, value):
        self.text = value
        self.save()
        return value

    @property
    def form_field(self):
        widget = (
            forms.Textarea
            if settings.CONTENT_BLOCKS_TEXTFIELD_TEXTAREA
            else forms.TextInput
        )
        # noinspection PyTypeChecker
        return forms.CharField(
            initial=self.text,
            required=self.template_field.required,
            help_text=self.template_field.help_text,
            widget=widget,
            max_length=256,
        )


class ContentField(ContentBlockField):
    class Meta:
        proxy = True

    @property
    def context_value(self):
        if settings.CONTENT_BLOCKS_MARK_SAFE:
            return mark_safe(self.content)
        return self.content

    def save_value(self, value):
        self.content = value
        self.save()
        return value

    @property
    def form_field(self):
        # noinspection PyTypeChecker
        return forms.CharField(
            initial=self.content,
            required=self.template_field.required,
            help_text=self.template_field.help_text,
            widget=forms.Textarea(),
        )


class CheckboxField(ContentBlockField):
    template_name = "content_blocks/partials/fields/checkbox.html"

    class Meta:
        proxy = True

    @property
    def context_value(self):
        return self.checkbox

    def save_value(self, value):
        self.checkbox = value
        self.save()
        return value

    @property
    def form_field(self):
        # noinspection PyTypeChecker
        return forms.BooleanField(
            initial=self.checkbox,
            required=False,
            help_text=self.template_field.help_text,
        )


@cleanup_ignore
class ImageField(ContentBlockField):
    preview_template_name = "content_blocks/partials/fields/previews/image.html"

    class Meta:
        proxy = True

    @property
    def context_value(self):
        return self.image

    def save_value(self, value):
        self.image = None if value is False else value
        self.save()
        return value

    @property
    def form_field(self):
        # noinspection PyTypeChecker
        return SVGAndImageFieldFormField(
            initial=self.image,
            required=self.template_field.required,
            help_text=self.template_field.help_text,
            widget=FileWidget(),
        )


@cleanup_ignore
class FileField(ContentBlockField):
    class Meta:
        proxy = True

    @property
    def context_value(self):
        return self.file

    def save_value(self, value):
        self.file = None if value is False else value
        self.save()
        return value

    @property
    def form_field(self):
        # noinspection PyTypeChecker
        return forms.FileField(
            initial=self.file,
            required=self.template_field.required,
            help_text=self.template_field.help_text,
            widget=FileWidget(),
        )


@cleanup_ignore
class VideoField(ContentBlockField):
    preview_template_name = "content_blocks/partials/fields/previews/video.html"

    class Meta:
        proxy = True

    @property
    def context_value(self):
        return self.video

    def save_value(self, value):
        self.video = None if value is False else value
        self.save()
        return value

    @property
    def form_field(self):
        # noinspection PyTypeChecker
        return forms.FileField(
            initial=self.video,
            required=self.template_field.required,
            help_text=self.template_field.help_text,
            widget=FileWidget(),
        )


class EmbeddedVideoField(ContentBlockField):
    preview_template_name = (
        "content_blocks/partials/fields/previews/embedded_video.html"
    )

    class Meta:
        proxy = True

    @property
    def context_value(self):
        return self.embedded_video

    def save_value(self, value):
        self.embedded_video = value
        self.save()
        return value

    @property
    def form_field(self):
        # noinspection PyTypeChecker
        return forms.CharField(
            initial=self.embedded_video,
            required=self.template_field.required,
            help_text=self.template_field.help_text,
            max_length=256,
        )


class IframeField(ContentBlockField):
    preview_template_name = "content_blocks/partials/fields/previews/iframe.html"

    class Meta:
        proxy = True

    @property
    def context_value(self):
        return self.iframe

    def save_value(self, value):
        self.iframe = value
        self.save()
        return value

    @property
    def form_field(self):
        # noinspection PyTypeChecker
        return forms.CharField(
            initial=self.iframe,
            required=self.template_field.required,
            help_text=self.template_field.help_text,
            max_length=256,
        )


class ChoiceField(ContentBlockField):
    class Meta:
        proxy = True

    @property
    def context_value(self):
        return self.choice

    def save_value(self, value):
        self.choice = value
        self.save()
        return value

    @property
    def form_field(self):
        choices = [[None, "-" * 9]] + json.loads(self.template_field.choices)
        # noinspection PyTypeChecker
        return forms.CharField(
            initial=self.choice,
            required=self.template_field.required,
            help_text=self.template_field.help_text,
            widget=forms.Select(choices=choices),
        )


class ModelChoiceField(ContentBlockField):
    class Meta:
        proxy = True

    @property
    def form_field(self):
        # self.model_choice_content_type must be set or this will raise an error.
        model = self.model_choice_content_type.model_class()
        queryset = model.objects.all()
        # noinspection PyTypeChecker
        return forms.ModelChoiceField(
            queryset,
            initial=self.model_choice,
            required=self.template_field.required,
            help_text=self.template_field.help_text,
        )

    @property
    def context_value(self):
        return self.model_choice

    def save_value(self, value):
        # If value is None, and we set self.model_choice = None it clears model_choice_content_type - not ideal
        # So instead we ste the model_choice_object_id
        self.model_choice_object_id = value.id if value is not None else None
        self.save()
        return value


class NestedField(ContentBlockField):
    class Meta:
        proxy = True

    @property
    def form_field(self):
        return None

    def save_value(self, value):
        return None

    @property
    def label(self):
        return pretty_name(self.template_field.key)

    @property
    def context_value(self):
        """
        :return: Nested content blocks which we can then call render or access context on.
        """
        return self.content_blocks.nested()


def optimise_queryset(func):
    """
    Decorator to optimise queryset.
    Moved from ContentBlockManager.get_queryset to here such that dumpdata can run without the select_related which
    causes it to fail when using natural_keys.
    """

    @functools.wraps(func)
    def wrapper(self):
        return (
            func(self)
            .prefetch_related("content_block_fields")
            .select_related("content_block_template")
        )

    return wrapper


class ContentBlockManager(VisibleManager):
    # todo consider moving some of these to service classes. Only need to keep those used in templates here?
    #   Could also take more care with optimisation and only apply it when needed.

    @optimise_queryset
    def visible(self):
        """
        Visible published only.
        Used in templates to render published content blocks.
        """
        return super().visible().filter(draft=False)

    @optimise_queryset
    def previews(self):
        """
        Visible drafts only. Exclude those which haven't been saved via the editor yet (empties).
        Can be used in page previews.
        """
        return super().visible().filter(draft=True, saved=True)

    @optimise_queryset
    def nested(self):
        """
        Used in the context for NestedField objects. Visible but with unsaved excluded.
        """
        return super().visible().filter(draft=False, saved=True)

    @optimise_queryset
    def published(self):
        """
        All published including visible=False.
        Used by the publishing/reset mechanism.
        """
        return self.get_queryset().filter(draft=False)

    @optimise_queryset
    def drafts(self):
        """
        All drafts including visible=False.
        Used by the editor and publishing/reset mechanism.
        """
        return self.get_queryset().filter(draft=True)


class ContentBlock(PositionModel, AutoDateModel, VisibleModel, CloneMixin):
    """
    Content Block Model
    Content block fields point here.  Those created are based on the content block template.
    """

    objects = ContentBlockManager()

    content_block_template = models.ForeignKey(
        "content_blocks.ContentBlockTemplate", on_delete=models.CASCADE
    )
    parent = models.ForeignKey(
        ContentBlockField,
        on_delete=models.CASCADE,
        related_name="content_blocks",
        blank=True,
        null=True,
    )

    name = models.CharField(
        max_length=320,
        blank=True,
        help_text="Used to identify content blocks in the content block editor. Defaults to "
        "<content_block_template_name> #<content_block_id>",
    )

    css_class = models.CharField(max_length=64, blank=True)

    draft = models.BooleanField(blank=True, default=False)

    saved = models.BooleanField(blank=True, default=False)

    context_name = "content_block"

    @cached_property
    def template(self):
        """
        HTML template
        """
        return self.content_block_template.template or None

    @cached_property
    def fields(self):
        """
        Create/cache a dict of key to content block field objects.
        """
        nested_fields = {}
        fields = {}
        for field in self.content_block_fields.all():
            fields[field.template_field.key] = field
            if field.template_field.field_type == ContentBlockFields.NESTED_FIELD:
                nested_fields[field.template_field.key] = field

        self.__dict__["nested_fields"] = nested_fields
        return fields

    @cached_property
    def nested_fields(self):
        """
        Create/cache a dict of nested fields.  If self.fields has already been called this is already cached.
        """
        return {
            field.template_field.key: field
            for field in self.content_block_fields.filter(
                template_field__field_type=ContentBlockFields.NESTED_FIELD
            )
        }

    @cached_property
    def context(self):
        """
        Return dictionary of template context for this content block
        """
        context = {key: field.context_value for key, field in self.fields.items()}
        context["css_class"] = self.css_class
        return context

    @cached_property
    def can_render(self):
        """
        :return: True if the template exists.
        """
        if not self.template:
            return False
        try:
            get_template(self.template)
            return True
        except TemplateDoesNotExist:
            return False

    def render(self):
        """
        Render html for this block.
        This should only be called by the template and exists to support legacy projects.
        The {% render_content_block %} template tag should be used in preference.
        """
        from content_blocks.services.content_block import RenderServices

        return RenderServices.render_content_block(self)


class ContentBlockTemplateManager(VisibleManager):
    def get_by_natural_key(self, name):
        return self.get(name=name)


class ContentBlockTemplate(PositionModel, AutoDateModel, VisibleModel):
    """
    Content Block Template model.
    Used to define content block types.
    """

    objects = ContentBlockTemplateManager()

    name = models.CharField(max_length=256, unique=True)

    template_filename = models.CharField(max_length=256, blank=True)

    def __str__(self):
        return self.name

    def natural_key(self):
        return (self.name,)

    @property
    def template(self):
        if self.template_filename:
            return f"content_blocks/content_blocks/{self.template_filename}"


class ContentBlockTemplateFieldManager(models.Manager):
    def get_by_natural_key(self, key, content_block_template_name):
        return self.get(
            key=key, content_block_template__name=content_block_template_name
        )


class ContentBlockTemplateField(PositionModel):
    """
    Content Block Template Field model.
    Used to define content block types.
    """

    objects = ContentBlockTemplateFieldManager()

    content_block_template = models.ForeignKey(
        ContentBlockTemplate,
        on_delete=models.CASCADE,
        related_name="content_block_template_fields",
    )

    field_type = models.CharField(max_length=32, choices=ContentBlockFields.choices)

    key = models.SlugField(
        max_length=64,
        validators=[
            RegexValidator(
                "[a-z0-9_]+", "Lowercase letters, numbers and underscores only."
            )
        ],
        help_text="Must be unique to this content block template. Lowercase letters, numbers and underscores only.",
    )

    help_text = models.TextField(blank=True)
    required = models.BooleanField(blank=True, default=False)

    css_class = models.CharField(
        max_length=64,
        blank=True,
        help_text="Set a custom CSS class for this field in the editor.",
    )

    # Only used if field_type == NestedField
    nested_templates = models.ManyToManyField(
        "content_blocks.ContentBlockTemplate",
        blank=True,
        help_text="Choose the content block templates that can be used in this nested field.",
    )
    min_num = models.PositiveIntegerField(
        default=0, help_text="The minimum number of nested blocks allowed."
    )
    max_num = models.PositiveIntegerField(
        default=99, help_text="The maximum number of nested blocks allowed."
    )

    # Only used if field_type == ChoiceField
    choices = models.TextField(blank=True)

    # Only used if field_type == ModelChoiceField
    model_choice_content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, blank=True, null=True
    )

    class Meta(PositionModel.Meta):
        constraints = [
            models.UniqueConstraint(
                fields=["key", "content_block_template"],
                name="unique_key_content_block_template",
            )
        ]

    def __str__(self):
        return self.key or super().__str__()

    def natural_key(self):
        return (self.key,) + self.content_block_template.natural_key()

    natural_key.dependencies = ["content_blocks.contentblocktemplate"]


class ContentBlockAvailability(AutoDateModel):
    """
    Used to set which content block templates can be added to a model.
    """

    # null=True is required for admin list to work. blank=False because this is a required field.
    content_type = models.OneToOneField(
        ContentType, on_delete=models.CASCADE, null=True
    )

    content_block_templates = models.ManyToManyField(ContentBlockTemplate)

    class Meta:
        verbose_name_plural = "Content block availabilities"

    def __str__(self):
        return str(self.content_type)


class ContentBlockParentModel(models.Model):
    """
    Model to be inherited by parent objects which then have content blocks assigned.  E.g. imperium.Page
    """

    content_blocks = models.ManyToManyField(ContentBlock)

    class Meta:
        abstract = True

    def delete(self, using=None, keep_parents=False):
        self.content_blocks.all().delete()
        return super().delete(using=using, keep_parents=keep_parents)


class ContentBlockCollection(AutoDateModel, ContentBlockParentModel):
    """
    Simple collections of content blocks identified by slug.
    Use the content_blocks_collection template tag to render a content block collection in a template.
    """

    name = models.CharField(
        max_length=256, blank=True, help_text="For identification purposes only."
    )

    slug = models.SlugField(unique=True)

    class Meta:
        ordering = ["name"]

    def __str__(self):
        return self.name or self.slug



from io import StringIO

from django.core.management import BaseCommand

from content_blocks.services.content_block_template import ImportExportServices


class Command(BaseCommand):
    help = "Export content block templates to json."

    def handle(self, *args, **options):
        """
        Export ContentBlockTemplate and associated ContentBlockTemplateField to json via stdout.
        The output can be used for import_content_block_templates management command.
        """
        buffer = StringIO()
        ImportExportServices.export_content_block_templates(file_like=buffer)
        self.stdout.write(buffer.getvalue())



"""
Functions to import and export (from/to JSON)
"""
import itertools
from pathlib import Path

from django.core import serializers
from django.core.management import call_command
from django.db import transaction

from content_blocks.models import (
    ContentBlock,
    ContentBlockField,
    ContentBlockTemplate,
    ContentBlockTemplateField,
)
from content_blocks.signals import post_import


class ImportExportServices:
    """
    Service for import and export.
    """

    # todo tests

    # Export ContentBlockTemplate
    @staticmethod
    def export_content_block_templates(queryset=None, file_like=None):
        """
        Export ContentBlockTemplate and related ContentBlockTemplateField to json.
        :param queryset: Queryset of ContentBlockTemplate defaults to ContentBlockTemplate.objects.all()
        :param file_like: Any file like object including HttpResponse and StringIO. This is where the serializer will
        write to.
        :return: Serializer value.
        """
        content_block_templates = queryset or ContentBlockTemplate.objects.all()
        content_block_template_fields = ContentBlockTemplateField.objects.filter(
            content_block_template__in=content_block_templates
        )

        return serializers.serialize(
            "json",
            itertools.chain(content_block_templates, content_block_template_fields),
            stream=file_like,
            use_natural_foreign_keys=True,
            use_natural_primary_keys=True,
        )

    # Import ContentBlockTemplate
    @staticmethod
    def add_new_content_block_template_field(content_block_template_field):
        """
        Add ContentBlockField object for the new ContentBlockTemplateField.
        """
        content_blocks = ContentBlock.objects.filter(
            content_block_template=content_block_template_field.content_block_template
        )

        # Create a new field for existing content blocks
        for content_block in content_blocks:
            ContentBlockField.objects.create(
                template_field=content_block_template_field,
                content_block=content_block,
                field_type=content_block_template_field.field_type,
            )

    @staticmethod
    def delete_old_content_block_template_fields(imported_pks):
        """
        For each ContentBlockTemplate specified in the json import delete any related ContentBlockTemplateField
        that is not specified in the json.
        """
        ContentBlockTemplateField.objects.filter(
            content_block_template_id__in=imported_pks[ContentBlockTemplate],
        ).exclude(
            id__in=imported_pks[ContentBlockTemplateField],
        ).delete()

    @staticmethod
    def import_content_block_templates(stream_or_string, verbosity=0):
        """
        Takes a stream or string and imports it.  Syncs ContentBlockField for ContentBlockTemplate imported by
        deleting ContentBlockField which aren't in the imported data but are in the database. And by creating
        ContentBlockField for ContentBlockTemplateField which are in the imported data but aren't in the database.
        :param verbosity: verbosity option passed to called commands defaults to 0.
        :param stream_or_string:
        """
        imported_pks = {ContentBlockTemplate: [], ContentBlockTemplateField: []}

        with transaction.atomic():
            for obj in serializers.deserialize("json", stream_or_string):
                if obj.object.__class__ not in imported_pks.keys():
                    # Ignore any objects that aren't ContentBlockTemplate or ContentBlockTemplateField
                    continue

                created = obj.object.pk is None
                if not created:
                    # Double check because tests show the above is not always enough
                    model = type(obj.object)
                    created = not model.objects.filter(pk=obj.object.pk).exists()

                obj.save()

                if created and isinstance(obj.object, ContentBlockTemplateField):
                    ImportExportServices.add_new_content_block_template_field(
                        obj.object
                    )

                imported_pks[obj.object.__class__].append(obj.object.pk)

            ImportExportServices.delete_old_content_block_template_fields(imported_pks)

            # Reorder ContentBlockTemplate
            call_command(
                "reorder", "content_blocks.ContentBlockTemplate", verbosity=verbosity
            )

        post_import.send(ContentBlockTemplate)

    @staticmethod
    def import_content_block_templates_from_file(filepath, verbosity=0):
        """
        Takes a filepath and opens the file and imports it.
        :param filepath: String. The path is relative to
        """
        filepath = Path(filepath)

        with filepath.open() as file:
            ImportExportServices.import_content_block_templates(file, verbosity)



# Generated by Django 4.2 on 2023-04-27 12:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('content_blocks', '0002_alter_contentblock_css_class_and_more'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='contentblocktemplatefield',
            unique_together=set(),
        ),
        migrations.AlterField(
            model_name='contentblocktemplatefield',
            name='field_type',
            field=models.CharField(choices=[('TextField', 'Text Field'), ('ContentField', 'Content Field'), ('ImageField', 'Image Field'), ('NestedField', 'Nested Field'), ('CheckboxField', 'Checkbox Field'), ('ChoiceField', 'Choice Field'), ('ModelChoiceField', 'Model Choice Field'), ('FileField', 'File Field'), ('VideoField', 'Video Field'), ('EmbeddedVideoField', 'Embedded Video Field')], max_length=32),
        ),
        migrations.AddConstraint(
            model_name='contentblocktemplatefield',
            constraint=models.UniqueConstraint(fields=('key', 'content_block_template'), name='unique_key_content_block_template'),
        ),
    ]



# Generated by Django 4.2 on 2023-04-14 09:38

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('content_blocks', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='contentblock',
            name='css_class',
            field=models.CharField(blank=True, max_length=64),
        ),
        migrations.AlterField(
            model_name='contentblockfield',
            name='field_type',
            field=models.CharField(max_length=32),
        ),
        migrations.AlterField(
            model_name='contentblocktemplatefield',
            name='css_class',
            field=models.CharField(blank=True, help_text='Set a custom CSS class for this field in the editor.', max_length=64),
        ),
        migrations.AlterField(
            model_name='contentblocktemplatefield',
            name='field_type',
            field=models.CharField(choices=[('TextField', 'Text Field'), ('ContentField', 'Content Field'), ('ImageField', 'Image Field'), ('VideoField', 'Video Field'), ('FileField', 'File Field'), ('EmbeddedVideoField', 'Embedded Video Field'), ('NestedField', 'Nested Field'), ('ModelChoiceField', 'Model Choice Field'), ('ChoiceField', 'Choice Field'), ('CheckboxField', 'Checkbox Field')], max_length=32),
        ),
        migrations.AlterField(
            model_name='contentblocktemplatefield',
            name='key',
            field=models.SlugField(help_text='Must be unique to this content block template. Lowercase letters, numbers and underscores only.', max_length=64, validators=[django.core.validators.RegexValidator('[a-z0-9_]+', 'Lowercase letters, numbers and underscores only.')]),
        ),
    ]



"""
Content Blocks conf.py
"""
from django.conf import settings as django_settings


class ContentBlocksSettings:
    # Choose the storage backend to use for media files
    # Image, field and video storage settings are use instead of content blocks storage if set.
    CONTENT_BLOCKS_IMAGE_STORAGE = None
    CONTENT_BLOCKS_VIDEO_STORAGE = None
    CONTENT_BLOCKS_FILE_STORAGE = None
    # Use default storage if not set.
    # Support provided here for Django 3.2
    CONTENT_BLOCKS_STORAGE = getattr(django_settings, "STORAGES", {}).get(
        "default", {}
    ).get("BACKEND") or getattr(
        django_settings,
        "DEFAULT_FILE_STORAGE",
        "django.core.files.storage.FileSystemStorage",
    )

    # Set the default status message.  Can be a callable.
    CONTENT_BLOCKS_DEFAULT_STATUS_MESSAGE = ""

    # Mark context for TextField and ContentField as safe. Useful for allowing html in these fields.
    CONTENT_BLOCKS_MARK_SAFE = False

    # Use a textarea widget for TextFields.
    CONTENT_BLOCKS_TEXTFIELD_TEXTAREA = False

    # Pre render content blocks on publish.
    # Useful for populating the cache and/or pre generating django-lazy-srcset images.
    CONTENT_BLOCKS_PRE_RENDER = True
    # Set the cache_timeout when pre rendering
    CONTENT_BLOCKS_PRE_RENDER_CACHE_TIMEOUT = None

    def __getattribute__(self, name):
        try:
            return getattr(django_settings, name)
        except AttributeError:
            return super().__getattribute__(name)


settings = ContentBlocksSettings()



"""
Content Blocks abstract_models.py
"""

from django.db import models


class VisibleManager(models.Manager):
    def visible(self):
        return self.get_queryset().filter(visible=True)


class VisibleModel(models.Model):
    visible = models.BooleanField(default=True, help_text="Uncheck to hide this.")

    class Meta:
        abstract = True


class AutoDateModel(models.Model):
    create_date = models.DateTimeField("Creation Date", auto_now_add=True)
    mod_date = models.DateTimeField("Last Modified", auto_now=True)

    class Meta:
        abstract = True
        ordering = ["-create_date"]


class PositionModel(models.Model):
    position = models.PositiveIntegerField(
        default=0, help_text="Set a custom ordering. Lower numbers appear first."
    )

    class Meta:
        abstract = True
        ordering = ["position"]
